@@ -33,10 +33,11 @@ function for information about how liblouis searches for these tables.
 @author: Eitan Isaacson <eitan@ascender.com>
 @author: Michael Whapples <mwhapples@aim.com>
 @author: Davy Kager <mail@davykager.nl>
+@author: Leonard de Ruijter (Babbage B.V.) <alderuijter@gmail.com>
 """
 
+from __future__ import unicode_literals
 from ctypes import *
-import struct
 import atexit
 import sys
 
@@ -49,28 +50,21 @@ def _createTypeformbuf(length, typeform=None):
     """Creates a typeform buffer for liblouis calls"""
     return (c_ushort*length)(*typeform) if typeform else (c_ushort*length)()
 
-createStr = None
-if sys.version_info[0] == 2:
-    createStr = lambda x: unicode(x)
-else:
-    createStr = lambda x: str(x)
-
-#{ Module Configuration
-#: Specifies the number by which the input length should be multiplied
-#: to calculate the maximum output length.
-#: @type: int
-# This default will handle the case where every input character is
-# undefined in the translation table.
-outlenMultiplier = 4 + sizeof(c_wchar) * 2
-#}
+createEncodedByeString = None
+if sys.version_info.major == 2:
+    createEncodedByeString = lambda x: unicode(x).encode(conversionEncoding)
+else: # sys.version_info[0] == 3
+    createEncodedByeString = lambda x: str(x).encode(conversionEncoding)
 
 try:
     # Native win32
     _loader = windll
+    _functype = WINFUNCTYPE
 except NameError:
     # Unix/Cygwin
     _loader = cdll
-liblouis = _loader["liblouis.so.16"]
+    _functype = CFUNCTYPE
+liblouis = _loader["liblouis.so.17"]
 
 atexit.register(liblouis.lou_free)
 
@@ -79,25 +73,25 @@ liblouis.lou_version.restype = c_char_p
 liblouis.lou_charSize.restype = c_int
 
 liblouis.lou_translateString.argtypes = (
-    c_char_p, c_wchar_p, POINTER(c_int), c_wchar_p, 
+    c_char_p, POINTER(c_char), POINTER(c_int), POINTER(c_char), 
          POINTER(c_int), POINTER(c_ushort), POINTER(c_char), c_int)
 
 liblouis.lou_translate.argtypes = (
-    c_char_p, c_wchar_p, POINTER(c_int), c_wchar_p, 
+    c_char_p, POINTER(c_char), POINTER(c_int), POINTER(c_char), 
          POINTER(c_int), POINTER(c_ushort), POINTER(c_char), 
          POINTER(c_int), POINTER(c_int), POINTER(c_int), c_int)
 
 liblouis.lou_backTranslateString.argtypes = (
-         c_char_p, c_wchar_p, POINTER(c_int), c_wchar_p,
+         c_char_p, POINTER(c_char), POINTER(c_int), POINTER(c_char),
          POINTER(c_int), POINTER(c_ushort), POINTER(c_char), c_int)
 
 liblouis.lou_backTranslate.argtypes = (
-         c_char_p, c_wchar_p, POINTER(c_int), c_wchar_p, POINTER(c_int),
+         c_char_p, POINTER(c_char), POINTER(c_int), POINTER(c_char), POINTER(c_int),
          POINTER(c_ushort), POINTER(c_char), POINTER(c_int), POINTER(c_int),
          POINTER(c_int), c_int)
 
 liblouis.lou_hyphenate.argtypes = (
-         c_char_p, c_wchar_p, c_int, POINTER(c_char), c_int)
+         c_char_p, POINTER(c_char), c_int, POINTER(c_char), c_int)
 
 liblouis.lou_checkTable.argtypes = (c_char_p,)
 
@@ -105,6 +99,39 @@ liblouis.lou_compileString.argtypes = (c_char_p, c_char_p)
 
 liblouis.lou_getTypeformForEmphClass.argtypes = (c_char_p, c_char_p)
 
+liblouis.lou_dotsToChar.argtypes = (
+    c_char_p, POINTER(c_char), POINTER(c_char), 
+         c_int, c_int)
+
+liblouis.lou_charToDots.argtypes = (
+    c_char_p, POINTER(c_char), POINTER(c_char), 
+         c_int, c_int)
+
+logcallback = _functype(None, c_int, c_char_p)
+
+liblouis.lou_registerLogCallback.restype = None
+liblouis.lou_registerLogCallback.argtypes = (logcallback,)
+
+liblouis.lou_setLogLevel.restype = None
+liblouis.lou_setLogLevel.argtypes = (c_int,)
+
+#{ Module Configuration
+#: Specifies the charSize (in bytes) used by liblouis.
+#: This is fetched once using L{liblouis.lou_charSize}.
+#: Call it directly, since L{charSize} is not yet defined.
+#: @type: int
+wideCharBytes = liblouis.lou_charSize()
+#: Specifies the number by which the input length should be multiplied
+#: to calculate the maximum output length.
+#: @type: int
+# This default will handle the case where every input character is
+# undefined in the translation table.
+outlenMultiplier = 4 + wideCharBytes * 2
+#: Specifies the encoding to use when converting from byte strings to unicode strings.
+#: @type: str
+conversionEncoding = "utf_%d_le" % (wideCharBytes * 8)
+#}
+
 def version():
     """Obtain version information for liblouis.
     @return: The version of liblouis, plus other information, such as
@@ -142,10 +169,10 @@ def translate(tableList, inbuf, typeform=None, cursorPos=0, mode=0):
     @see: lou_translate in the liblouis documentation
     """
     tablesString = _createTablesString(tableList)
-    inbuf = createStr(inbuf)
-    inlen = c_int(len(inbuf))
+    inbuf = createEncodedByeString(inbuf)
+    inlen = c_int(len(inbuf) // wideCharBytes)
     outlen = c_int(inlen.value*outlenMultiplier)
-    outbuf = create_unicode_buffer(outlen.value)
+    outbuf = create_string_buffer(outlen.value * wideCharBytes)
     typeformbuf = None
     if typeform:
         typeformbuf = _createTypeformbuf(outlen.value, typeform)
@@ -158,7 +185,7 @@ def translate(tableList, inbuf, typeform=None, cursorPos=0, mode=0):
         raise RuntimeError("Can't translate: tables %s, inbuf %s, typeform %s, cursorPos %s, mode %s"%(tableList, inbuf, typeform, cursorPos, mode))
     if isinstance(typeform, list):
         typeform[:] = list(typeformbuf)
-    return outbuf.value, inPos[:outlen.value], outPos[:inlen.value], cursorPos.value
+    return outbuf.raw[:outlen.value * wideCharBytes].decode(conversionEncoding), inPos[:outlen.value], outPos[:inlen.value], cursorPos.value
 
 def translateString(tableList, inbuf, typeform=None, mode=0):
     """Translate a string of characters.
@@ -177,10 +204,10 @@ def translateString(tableList, inbuf, typeform=None, mode=0):
     @see: lou_translateString in the liblouis documentation
     """
     tablesString = _createTablesString(tableList)
-    inbuf = createStr(inbuf)
-    inlen = c_int(len(inbuf))
+    inbuf = createEncodedByeString(inbuf)
+    inlen = c_int(len(inbuf) // wideCharBytes)
     outlen = c_int(inlen.value*outlenMultiplier)
-    outbuf = create_unicode_buffer(outlen.value)
+    outbuf = create_string_buffer(outlen.value * wideCharBytes)
     typeformbuf = None
     if typeform:
         typeformbuf = _createTypeformbuf(outlen.value, typeform)
@@ -190,7 +217,7 @@ def translateString(tableList, inbuf, typeform=None, mode=0):
         raise RuntimeError("Can't translate: tables %s, inbuf %s, typeform %s, mode %s"%(tableList, inbuf, typeform, mode))
     if isinstance(typeform, list):
         typeform[:] = list(typeformbuf)
-    return outbuf.value
+    return outbuf.raw[:outlen.value * wideCharBytes].decode(conversionEncoding)
 
 def backTranslate(tableList, inbuf, typeform=None, cursorPos=0, mode=0):
     """Back translates a string of characters, providing position information.
@@ -213,10 +240,10 @@ def backTranslate(tableList, inbuf, typeform=None, cursorPos=0, mode=0):
     @see: lou_backTranslate in the liblouis documentation.
     """
     tablesString = _createTablesString(tableList)
-    inbuf = createStr(inbuf)
-    inlen = c_int(len(inbuf))
+    inbuf = createEncodedByeString(inbuf)
+    inlen = c_int(len(inbuf) // wideCharBytes)
     outlen = c_int(inlen.value * outlenMultiplier)
-    outbuf = create_unicode_buffer(outlen.value)
+    outbuf = create_string_buffer(outlen.value * wideCharBytes)
     typeformbuf = None
     if isinstance(typeform, list):
         typeformbuf = _createTypeformbuf(outlen.value)
@@ -229,7 +256,7 @@ def backTranslate(tableList, inbuf, typeform=None, cursorPos=0, mode=0):
         raise RuntimeError("Can't back translate: tables %s, inbuf %s, typeform %s, cursorPos %d, mode %d"%(tableList, inbuf, typeform, cursorPos, mode))
     if isinstance(typeform, list):
         typeform[:] = list(typeformbuf)
-    return outbuf.value, inPos[:outlen.value], outPos[:inlen.value], cursorPos.value
+    return outbuf.raw[:outlen.value * wideCharBytes].decode(conversionEncoding), inPos[:outlen.value], outPos[:inlen.value], cursorPos.value
 
 def backTranslateString(tableList, inbuf, typeform=None, mode=0):
     """Back translate from Braille.
@@ -248,10 +275,10 @@ def backTranslateString(tableList, inbuf, typeform=None, mode=0):
     @see: lou_backTranslateString in the liblouis documentation.
     """
     tablesString = _createTablesString(tableList)
-    inbuf = createStr(inbuf)
-    inlen = c_int(len(inbuf))
+    inbuf = createEncodedByeString(inbuf)
+    inlen = c_int(len(inbuf) // wideCharBytes)
     outlen = c_int(inlen.value * outlenMultiplier)
-    outbuf = create_unicode_buffer(outlen.value)
+    outbuf = create_string_buffer(outlen.value * wideCharBytes)
     typeformbuf = None
     if isinstance(typeform, list):
         typeformbuf = _createTypeformbuf(outlen.value)
@@ -260,7 +287,7 @@ def backTranslateString(tableList, inbuf, typeform=None, mode=0):
         raise RuntimeError("Can't back translate: tables %s, inbuf %s, mode %d"%(tableList, inbuf, mode))
     if isinstance(typeform, list):
         typeform[:] = list(typeformbuf)
-    return outbuf.value
+    return outbuf.raw[:outlen.value * wideCharBytes].decode(conversionEncoding)
 
 def hyphenate(tableList, inbuf, mode=0):
     """Get information for hyphenation.
@@ -281,8 +308,8 @@ def hyphenate(tableList, inbuf, mode=0):
     @see: lou_hyphenate in the liblouis documentation.
     """
     tablesString = _createTablesString(tableList)
-    inbuf = createStr(inbuf)
-    inlen = c_int(len(inbuf))
+    inbuf = createEncodedByeString(inbuf)
+    inlen = c_int(len(inbuf) // wideCharBytes)
     hyphen_string = create_string_buffer(inlen.value + 1) 
     if not liblouis.lou_hyphenate(tablesString, inbuf, inlen, hyphen_string, mode):
         raise RuntimeError("Can't hyphenate: tables %s, inbuf %s, mode %d"%(tableList, inbuf, mode))
@@ -327,11 +354,78 @@ def getTypeformForEmphClass(tableList, emphClass):
     tablesString = _createTablesString(tableList)
     return liblouis.lou_getTypeformForEmphClass(tablesString, emphClass)
 
+def dotsToChar(tableList, inbuf):
+    """"Convert a string of dot patterns to a string of characters according to the specifications in tableList.
+    @param tableList: A list of translation tables.
+    @type tableList: list of str
+    @param inbuf: a string of dot patterns, either in liblouis format or Unicode braille.
+    @type inbuf: str
+    @raise RuntimeError: If a complete conversion could not be done.
+    @see: lou_dotsToChar in the liblouis documentation
+    """
+    tablesString = _createTablesString(tableList)
+    inbuf = createEncodedByeString(inbuf)
+    length = c_int(len(inbuf) // wideCharBytes)
+    outbuf = create_string_buffer(length.value * wideCharBytes)
+    if not liblouis.lou_dotsToChar(tablesString, inbuf, outbuf, length, 0):
+        raise RuntimeError("Can't convert dots to char: tables %s, inbuf %s"%(tableList, inbuf))
+    return outbuf.raw[:length.value * wideCharBytes].decode(conversionEncoding)
+
+def charToDots(tableList, inbuf, mode=0):
+    """"Convert a string of characterss to a string of dot patterns according to the specifications in tableList.
+    @param tableList: A list of translation tables.
+    @type tableList: list of str
+    @param inbuf: a string of characters.
+    @type inbuf: str
+    @param mode: The translation mode; add multiple values for a combined mode.
+    @type mode: int
+    @raise RuntimeError: If a complete conversion could not be done.
+    @see: lou_charToDOts in the liblouis documentation
+    """
+    tablesString = _createTablesString(tableList)
+    inbuf = createEncodedByeString(inbuf)
+    length = c_int(len(inbuf) // wideCharBytes)
+    outbuf = create_string_buffer(length.value * wideCharBytes)
+    if not liblouis.lou_charToDots(tablesString, inbuf, outbuf, length, mode):
+        raise RuntimeError("Can't convert char to dots: tables %s, inbuf %s, mode %d"%(tableList, inbuf, mode))
+    return outbuf.raw[:length.value * wideCharBytes].decode(conversionEncoding)
+
+def registerLogCallback(logcallback):
+    """Register logging callbacks.
+    Set to C{None} for default callback.
+    @param logcallback: The callback to use.
+        The callback must take two arguments:
+        @param level: The log level on which a message is logged.
+        @type level: int
+        @param message: The logged message.
+            Note that the callback should provide its own ASCII decoding routine.
+        @type message: bytes
+
+        Example callback:
+
+        @louis.logcallback
+        def incomingLouisLog(level, message):
+            print("Message %s logged at level %d" % (message.decode("ASCII"), level))
+
+    @type logcallback: L{logcallback}
+    """
+    return liblouis.lou_registerLogCallback(logcallback)
+
+def setLogLevel(level):
+	"""Set the level for logging callback to be called at.
+	@param level: one of the C{LOG_*} constants.
+	@type level: int
+	@raise ValueError: If an invalid log level is provided.
+	"""
+	if level not in logLevels:
+		raise ValueError("Level %d is an invalid log level"%level)
+	return liblouis.lou_setLogLevel(level)
+
 #{ Typeforms
 plain_text = 0x0000
-italic = 0x0001       # emph_1
-underline = 0x0002    # emph_2
-bold = 0x0004         # emph_3
+emph_1 = comp_emph_1 = italic = 0x0001
+emph_2 = comp_emph_2 = underline = 0x0002
+emph_3 = comp_emph_3 = bold = 0x0004
 emph_4 = 0x0008
 emph_5 = 0x0010
 emph_6 = 0x0020
@@ -354,10 +448,17 @@ noUndefinedDots = 128
 partialTrans = 256
 #}
 
-if charSize() == 2 and sys.maxunicode != 65535:
-    raise Exception("mismatch in Unicode width: liblouis is 2 and python isn't")
-elif charSize() == 4 and sys.maxunicode != 1114111:
-    raise Exception("mismatch in Unicode width: liblouis is 4 and python isn't")
+#{ logLevels
+LOG_ALL = 0
+LOG_DEBUG = 10000
+LOG_INFO = 20000
+LOG_WARN = 30000
+LOG_ERROR = 40000
+LOG_FATAL = 50000
+LOG_OFF = 60000
+#}
+
+logLevels = (LOG_ALL, LOG_DEBUG, LOG_INFO, LOG_WARN, LOG_ERROR, LOG_FATAL, LOG_OFF)
 
 if __name__ == '__main__':
     # Just some common tests.
